import java.io.DataInputStream;
import java.io.StringBufferInputStream;
import java.util.List;
import java.util.Vector;
import java.lang.Character;
import java.io.IOException;
import java.lang.StringBuffer;

// This class is a scanner for html. Input is an html file. Output is a list of
// HtmlElement. Each HtmlElement corresponds to a tag or text. Eg. 
// <a href=foo>foo</a> would result in 3 HtmlElements.
public class HtmlScanner 
{
    private static boolean debug = false;
    private Tags _tags;

    public HtmlScanner(Tags tags)
    {
        _tags = tags;
    }

    private void dbgprint( String str )
    {
        if ( debug ) {
            System.out.println( str );
        }
    }

    public void parse( DataInputStream input )
    {
        prepare();
        _elements = new Vector();

        try {
            for ( ;; ) {
                int ret = input.read();
                if ( ret == -1 ) {
                    break;
                }

                parseCharacter( (char)ret );
            }
        } catch ( IOException e ) {
            System.out.println( e.toString() );
        }

        cleanUp();
    }

    public List getElements()
    {
        return _elements;
    }

    private void prepare()
    {
        cleanUp();
    }

    private void cleanUp()
    {
        if ( _potentialTag != null ) {
            processTag();
            _potentialTag = null;
        }
    }

    private static final int PARSING_NOTHING = 0;
    private static final int PARSING_TEXT = 1;
    private static final int PARSING_TAG_NAME = 2;
    private static final int PARSING_TAG_ATTRIBUTE = 3;
    private static final int PARSING_TAG_ATTRIBUTE_NAME = 4;
    private static final int PARSING_TAG_ATTRIBUTE_NAME_OR_VALUE = 5;
    private static final int PARSING_TAG_ATTRIBUTE_VALUE = 6;
    private static final int PARSING_COMMENT = 7;
    private static final int PARSING_ENDTAG = 8;
    private static final int PARSING_SCRIPT = 9;

    private String stateName(int state )
    {
        return state == PARSING_NOTHING ? "PARSING_NOTHING" :
            state == PARSING_TEXT ? "PARSING_TEXT" :
            state == PARSING_TAG_NAME ? "PARSING_TAG_NAME" :
            state == PARSING_TAG_ATTRIBUTE ? "PARSING_TAG_ATTRIBUTE":
            state == PARSING_TAG_ATTRIBUTE_NAME ? "PARSING_TAG_ATTRIBUTE_NAME" :
            state == PARSING_TAG_ATTRIBUTE_NAME_OR_VALUE ? "PARSING_TAG_ATTRIBUTE_NAME_OR_VALUE" :
            state == PARSING_TAG_ATTRIBUTE_VALUE ? "PARSING_TAG_ATTRIBUTE_VALUE" :
            state == PARSING_COMMENT ? "PARSING_COMMENT" :
            state == PARSING_SCRIPT ? "PARSING_SCRIPT" :
            "UNKNOWN";
    }

    private static final int MAX_TAG_NAME_LENGTH = 32;
    private static final int MAX_TAG_ATTRIBUTE_NAME_LENGTH = 32;
    private static final int MAX_TAG_ATTRIBUTE_VALUE_LENGTH = 4096;

    private void parseCharacter( char ch )
    {
        dbgprint( "Parse char: \'" + ch + "\'");
        switch( ch )
        {
            case '<':
                beginNewTag();
                break;

            case '>':
                endBeginTag();
                break;

            default:
                switch( _parseState )
                {
                    case PARSING_TAG_NAME:
                        parseTagName( ch );
                        break;
                    case PARSING_TAG_ATTRIBUTE_NAME:
                        parseTagAttributeName( ch );
                        break;
                    case PARSING_TAG_ATTRIBUTE_NAME_OR_VALUE:
                        parseTagAttributeNameOrValue( ch );
                        break;
                    case PARSING_TAG_ATTRIBUTE_VALUE:
                        parseTagAttributeValue( ch );
                        break;
                    case PARSING_COMMENT:
                    case PARSING_TEXT:
                        parseText( ch );
                        break;

                    case PARSING_SCRIPT:
                        parseScript( ch );
                        break;
                }
                break;
        }
    }

    private void beginNewTag()
    {
        if ( _potentialTag != null ) {
            _potentialTag = null;
        }

        if ( _text != null && _text.length() > 0  ) {
            addElement( new HtmlText(_text.toString()) );
            _text = null;
        }

        _potentialTag = new HtmlTag();
        beginParseState( PARSING_TAG_NAME );

        if ( _scriptBeginTag != null ) {
            _potentialScriptSuffix.append( '<' );
        }
    }

    private void endBeginTag()
    {
        switch ( _parseState ) {
            case PARSING_TAG_NAME:
                if ( !checkScriptEnd() ) {
                    processTag();
                    beginParseState(PARSING_TEXT);
                }
                break;
            case PARSING_TAG_ATTRIBUTE_NAME:
            case PARSING_TAG_ATTRIBUTE_NAME_OR_VALUE:
            case PARSING_TAG_ATTRIBUTE_VALUE:
                if ( _tagAttributeName.length() > 0  ){
                    if ( _tagAttributeValue != null ) {
                        _potentialTag.addAttribute(_tagAttributeName.toString(), 
                                _tagAttributeValue.toString());
                    } else {
                        _potentialTag.addAttribute(_tagAttributeName.toString(), 
                                null);
                    }
                }
                _tagAttributeName = new StringBuffer();
                _tagAttributeValue = new StringBuffer();
                beginParseState(PARSING_TEXT);
                processTag();
                break;

            case PARSING_SCRIPT:
                parseScript( '>' );
                break;

            case PARSING_TEXT:
                // not parsing a tag. Just add it to the document as text.
                // ParseText('>');
                break;
        }

    }

    private boolean checkScriptEnd()
    {
        if ( _scriptBeginTag != null ) {
            dbgprint("Script begin tag: " + _scriptBeginTag + 
                    ", potentialTagName: " + _potentialTag.name.toString() );
            // We are parsing text following '<' in the middle of a script.
            if ( _potentialTag.name.toString().toLowerCase().equals( "/" + 
                        _scriptBeginTag ) ) 
            {
                // It turned out to be the end of the script tag after
                // all.
                processScript();
                _scriptBeginTag = null;
                beginParseState( PARSING_TAG_ATTRIBUTE_NAME );
            } else {
                // It was not really the end of the script yet.
                _script.append( _potentialScriptSuffix );
                dbgprint("Switch to state PARSING_SCRIPT, suffix was " +
                        _potentialScriptSuffix.toString());
                _potentialScriptSuffix = null;
                _parseState = PARSING_SCRIPT;
                return true;
            }
        }
        return false;
    }

    private void parseTagName( char ch )
    {
        if ( _scriptBeginTag != null ) {
            _potentialScriptSuffix.append( ch );
        }
        
        int curTagNameLength = _potentialTag.name.length();
        if ( curTagNameLength > MAX_TAG_NAME_LENGTH ) {
            // abort this tag. Tag name too long.
            _potentialTag = new HtmlTag();
            beginParseState(PARSING_TEXT);

            return;
        }

        if ( Character.isWhitespace( ch ) ) {
            if ( _foundPrintableChar ) 
            {
                // strlwr(potentialTag.name);
                if ( checkScriptEnd() ) {
                } else if ( _potentialTag.name.toString().equals( "!--" ) ) {
                    beginParseState( PARSING_COMMENT );
                } else {
                    beginParseState( PARSING_TAG_ATTRIBUTE_NAME );
                }
            }
        } else {
            _potentialTag.name.append( ch );
            _foundPrintableChar = true;
        }
    }

    private void parseTagAttributeName( char ch )
    {
        if ( _tagAttributeName.length() > MAX_TAG_ATTRIBUTE_NAME_LENGTH ) {
            // too long. abort this attribute.
            _tagAttributeName = new StringBuffer();
            //todo: Should convert the entire tag to text element....
            return;
        }

        switch( ch ) {
            case '=':
                if ( _foundPrintableChar ) {
                    // Whitespace char. Go to next state.
                    //TrimString( szTagAttributeName
                    // strlwr(szTagAttributeName)
                    _tagAttributeName = 
                        new StringBuffer( _tagAttributeName.toString().trim().toLowerCase() );
                    beginParseState( PARSING_TAG_ATTRIBUTE_VALUE );
                } else {
                    // ERROR! No tag name!
                    // abort the tag
                    beginParseState( PARSING_TEXT );
                }
                break;

            default:
                _tagAttributeName.append( ch );
                _foundPrintableChar = true;
        }
    }

    private void parseTagAttributeValue( char ch )
    {
        if ( _tagAttributeValue.length() > MAX_TAG_ATTRIBUTE_VALUE_LENGTH ) {
            // Too long . abort entire attribute...
            // Todo: convert to text...
            _tagAttributeValue = new StringBuffer();
            _tagAttributeName = new StringBuffer();
            beginParseState( PARSING_TAG_ATTRIBUTE_NAME );
            return;
        }

        if ( Character.isWhitespace( ch ) ) {
            if ( !_tagAttributeValueInQuotes && _foundPrintableChar ) {
                //Whitespace char. Go to next sstate.
                _potentialTag.addAttribute(_tagAttributeName.toString(), 
                        _tagAttributeValue.toString() );
                _tagAttributeName = new StringBuffer();
                _tagAttributeValue = new StringBuffer();
                beginParseState( PARSING_TAG_ATTRIBUTE_NAME );
            } else if ( _tagAttributeValueInQuotes && _foundPrintableChar ) {
                _tagAttributeValue.append( ch );
            }
        } else if ( (ch == '\'' || ch == '\"' ) ) {
            if ( !_foundPrintableChar && !_tagAttributeValueInQuotes ) {
                _tagAttributeValueInQuotes = true;
                _foundPrintableChar = true;
                _quoteChar = ch;
            } else if ( _quoteChar == ch ) {
                _tagAttributeValueInQuotes = false;
            } else {
                _tagAttributeValue.append( ch );
            }
            _foundPrintableChar = true;
        } else {
            _tagAttributeValue.append( ch );
            _foundPrintableChar = true;
        }
    }

    private void parseTagAttributeNameOrValue( char ch )
    {

    }
     
    private void parseText( char ch )
    {
        _text.append( ch );
    }

    private void parseScript( char ch )
    {
        _script.append( ch );
    }

    private void processScript()
    {
        String type = "js";
        if ( _scriptBeginTag.toString().equals("style") ) {
            type = "css";
        }
        JsScanner js = new JsScanner(_tags, type);
        addElement( new JsElement( js.parse( new DataInputStream( new StringBufferInputStream( _script.toString() ) ) ), type ) );
        
        _script = null;
        _scriptBeginTag = null;
        _potentialScriptSuffix = null;
    }

    private void processTag()
    {
        // lower the tag name.
        _potentialTag.name = 
            new StringBuffer(_potentialTag.name.toString().toLowerCase());

        if ( _potentialTag.name.toString().equals("script") ||
                _potentialTag.name.toString().equals("style") ) {
            _scriptBeginTag = _potentialTag.name.toString();
            dbgprint("Script begin tag: " + _scriptBeginTag ); 
            beginParseState( PARSING_SCRIPT );
        }

        addElement( _potentialTag );
        _potentialTag = null;
    }

    private void beginParseState( int newParseState )
    {
        _foundPrintableChar = false;
        _tagAttributeValueInQuotes = false;
        _parseState = newParseState;
        dbgprint("Switch to state " + stateName(newParseState) );

        switch( _parseState ) {
            case PARSING_TAG_NAME:
                _potentialScriptSuffix = new StringBuffer();
                _potentialTag = new HtmlTag();
                break;
            case PARSING_TAG_ATTRIBUTE_NAME:
                _tagAttributeName = new StringBuffer();
                break;

             case PARSING_TAG_ATTRIBUTE_NAME_OR_VALUE:
                break;

             case PARSING_TAG_ATTRIBUTE_VALUE:
                _tagAttributeValue = new StringBuffer();
                break;

             case PARSING_SCRIPT:
                _script = new StringBuffer();
                break;

             case PARSING_COMMENT:
             case PARSING_TEXT:
                _text = new StringBuffer();
                break;
        }
    }

    private void addElement( HtmlElement element )
    {
        _elements.add( element );
    }

    private int _parseState = PARSING_NOTHING;
    private HtmlTag _potentialTag;
    private StringBuffer _tagAttributeName;
    private StringBuffer _tagAttributeValue;
    private boolean _foundPrintableChar;
    private boolean _tagAttributeValueInQuotes;
    private char _quoteChar;
    private StringBuffer _text;
    private String _scriptBeginTag;

	// The script
	private StringBuffer _script;

	// During parsing of a script, we may encounter "<". After that point we
	// put the characters into here. If it doesn't end up begin "</script" then
	// we will add the suffix to the script and go back to parsing script.
	private StringBuffer _potentialScriptSuffix;
    private List _elements;
}
